// BLED112 advertisement application
// Written by Gabor Paller <gaborpaller@gmail.com>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>

#include "cmd_def.h"

//#define DEBUG 0
// From attributes.txt, generated by Bluegiga's BGBUILD script. This is the default for BLED112, can be
// different for other modules
#define CTS_2A2B 17

int volatile serial_handle;
int volatile disconnect_issued;

void hexdump(uint8* data, int len ) {
	int i = 0,n;
	uint8* p = data;
	for( n = 0 ; n < len ; ++n ) {
	   printf( "%02x ",*p );
	   ++p;
	   ++i;
	   if( i >= 16 ) {
		printf("\n");
		i=0;
	   }
	}
	if( i > 0 )
		printf("\n");
}

void print_bdaddr(bd_addr bdaddr)
{
    printf("%02x:%02x:%02x:%02x:%02x:%02x",
            bdaddr.addr[5],
            bdaddr.addr[4],
            bdaddr.addr[3],
            bdaddr.addr[2],
            bdaddr.addr[1],
            bdaddr.addr[0]);
}

int readbuf(uint8* buf,int len ) {
	int i,dataread,retryCtr;
	uint8	c;
	uint8	*p = buf;
	retryCtr = 0;
	for( i=0 ; i < len ; ++i ) {
		do {
			dataread = read(serial_handle,&c,1);
			if( dataread < 0 ) {
				if( errno == EAGAIN ) {
					++retryCtr;
					if( retryCtr <= 3 ) {
						usleep( 100*1000 );	// 100 msec
						continue;
					}
				}
				if( errno != EAGAIN ) {
					char *errmsg = strerror(errno);
					printf( "readbuf error: %s\n",errmsg );
				}
				return -1;
			}
			if( dataread > 0 )
				break;
		} while( 1 );
//#ifdef DEBUG
//		printf( "readbuf[%d]: %02x\n",i,c);
//#endif
		*p = c;
		p++;
	}
	return len;
}

int read_response(int len) {
	uint8	buf[128];
	int rl;

#ifdef DEBUG
	printf( "Expecting %d bytes response\n",len);
#endif
	rl=readbuf(buf,len);
#ifdef DEBUG
	printf( "Data read: %d\n",rl);
    	hexdump(buf,rl);
#endif
	return rl == len ? 1 : 0;
}

void output(uint8 len1,uint8* data1,uint16 len2,uint8* data2)
{
    int written;

#ifdef DEBUG
    printf( "output: data1: %d\n",len1);
    hexdump(data1,len1);
#endif
    written = write(serial_handle,data1,len1);
    if( written < 0 ) {
	char *errormsg = strerror(errno);
	printf("ERROR: Writing data (1). %s\n",errormsg);
	exit(-1);
    }

    if( len2 != 0 ) {
#ifdef DEBUG
      printf( "output: data2: %d\n",len2);
      hexdump(data2,len2);
#endif
      written = write(serial_handle,data2,len2);
      if( written < 0 ) {
	char *errormsg = strerror(errno);
	printf("ERROR: Writing data (2). %s\n",errormsg);
	exit(-1);
      }
    }
}

int read_message()
{
    struct ble_header apihdr;
    unsigned char data[256];//enough for BLE
    int dataread;
    const struct ble_msg *apimsg;

#ifdef DEBUG
    printf("read1\n");
#endif
    dataread = readbuf((uint8*)&apihdr,4);
#ifdef DEBUG
    printf( "Data read (1): %d\n",dataread);
#endif
    if( dataread <= 0 ) {
	return 1;
    }
#ifdef DEBUG
    printf("data read1\n");
    hexdump((uint8*)&apihdr,dataread);
#endif
    if(apihdr.lolen) {
#ifdef DEBUG
    	printf("read2\n");
#endif
	dataread = readbuf(data,apihdr.lolen);
	if( dataread < 0 ) {
		return 1;
	}
#ifdef DEBUG
    	printf( "Data read (2): %d\n",dataread);
    	printf("data read2\n");
    	hexdump(data,dataread);
#endif
    }
    apimsg=ble_get_msg_hdr(apihdr);
    if(!apimsg) {
        printf("ERROR: Message not found:%d:%d\n",(int)apihdr.cls,(int)apihdr.command);
        return -1;
    }
    apimsg->handler(data);
    return 0;
}

void get_time(struct tm *dtm) {
	time_t t;
	struct tm tm;

	t = time(NULL);
	tm = *localtime(&t);
	memcpy(dtm,&tm,sizeof(struct tm));
}

void set_adv_time(int cts_handle,struct tm *advtm ) {
	uint8 cts_time[10];
	int v;
	
	memset(cts_time,0,10);
	v = advtm->tm_year+1900;
	cts_time[0] = (uint8)( v & 0xFF );
	cts_time[1] = (uint8)( ( v >> 8 ) & 0xFF );
	cts_time[2] = (uint8)( advtm->tm_mon+1 );
	cts_time[3] = (uint8)( advtm->tm_mday );
	cts_time[4] = (uint8)( advtm->tm_hour );
	cts_time[5] = (uint8)( advtm->tm_min );
	cts_time[6] = (uint8)( advtm->tm_sec );
	cts_time[7] = (uint8)( advtm->tm_wday+1 );
	ble_cmd_attributes_write(cts_handle,0,10,cts_time);
}

void ble_evt_gap_scan_response(const struct ble_msg_gap_scan_response_evt_t *msg)
{
    char name[32];
    int i,n;

    name[0]=0;
    // Get local name
    for (i = 0; i < msg->data.len; ) {
        int8 len = msg->data.data[i++];
        if (!len) continue;
        if (i + len > msg->data.len) break; // not enough data
        uint8 type = msg->data.data[i++];
        switch (type) {
        case 0x09:
            memcpy(name, msg->data.data + i, len - 1);
            name[len - 1] = '\0';
	    break;
        }

        i += len - 1;
    }

    print_bdaddr(msg->sender);
    printf(" RSSI: %d", msg->rssi);

    printf(" Name:");
    if (name[0]) printf("%s", name);
    else printf("Unknown");
    printf("\n");
    for (i = 0; i < msg->data.len; ) {
        int8 len = msg->data.data[i++];
        if (!len) continue;
        if (i + len > msg->data.len) break; // not enough data
        uint8 type = msg->data.data[i++];
        switch (type) {
        case 0x09:
	    break;
	default:
	    printf( "  field: 0x%x",type );
	    for( n = 1 ; n < len ; ++n )
		printf( " 0x%x",msg->data.data[i+n-1]);
	    printf("\n");
        }

        i += len - 1;
    }
}

void ble_evt_connection_status(const struct ble_msg_connection_status_evt_t *msg)
{
    if(msg->flags&connection_connected) {
        printf("#connected\n");
    } 
/*
    else {
        printf("#disconnected\n");
    // Start advertising
    	ble_cmd_gap_set_mode(1, 1);
    }
*/
}

void ble_evt_connection_disconnected(const struct ble_msg_connection_disconnected_evt_t *msg)
{
	printf("#disconnected\n");
    // Start advertising
        disconnect_issued = 1;
    	ble_cmd_gap_set_mode(1, 1);
}

void ble_rsp_gap_set_mode(const struct ble_msg_gap_set_mode_rsp_t *msg)
{
	printf( "#Mode set\n" );
	disconnect_issued = 0;
}


void ble_evt_system_protocol_error(const struct ble_msg_system_protocol_error_evt_t*msg)
{

	if( disconnect_issued ) {
		printf( "#Protocol error, gap_set_mode, repeated\n");
    		ble_cmd_gap_set_mode(1, 1);
	}
}

void print_help()
{
    printf("Demo application for BLE advertisement\n");
    printf("\tadv_example\t<com port> [<CTC characteristic handle>]\n");
    printf("\te.g. adv_example /dev/ttyACM0\n");
}

int main(int argc, char *argv[] )
{    
    char str[80];
    int status;
    struct termios settings;
    speed_t baud = B115200; /* baud rate */
    struct tm adv_time,current_time;
    int cts_handle = CTS_2A2B;


    if(argc<2)
    {
        print_help();
        exit(-1);
    }
    strcpy(str,argv[1]);
    if( argc == 3 ) {
	cts_handle = atoi(argv[2]);
	if( cts_handle == 0 ) {
		printf( "%s is not a valid CTS handle\n",argv[2]);
		exit(1);
	}
	printf( "CTS handle: %d\n",cts_handle);
    }
    serial_handle = open(str,O_RDWR|O_NONBLOCK);
    if( serial_handle < 0 ) {
	char *errmsg = strerror(errno);
        printf("Error opening serial port %s,%s\n",argv[1],errmsg);
	return -1;
    }
    tcgetattr(serial_handle, &settings);
    settings.c_cflag |= CLOCAL | CREAD;
    settings.c_cflag &= ~CSIZE;
    settings.c_cflag |= CS8;
    settings.c_cflag &= ~(CSTOPB);
    settings.c_cflag &= ~(PARENB|PARODD);
    settings.c_cflag &= ~(CRTSCTS);
    settings.c_lflag &= ~(ICANON|ECHO|ECHOE|ECHOK|ECHONL|ISIG|IEXTEN);
    settings.c_oflag &= ~(OPOST);
    settings.c_iflag &= ~(INLCR|IGNCR|ICRNL|IGNBRK);
    settings.c_iflag &= ~(INPCK|ISTRIP);
    settings.c_iflag &= ~(IXON|IXOFF);

    cfsetispeed(&settings, baud);
    cfsetospeed(&settings, baud); /* baud rate */
    tcsetattr(serial_handle, TCSANOW, &settings); /* apply the settings */
    tcflush(serial_handle,TCIOFLUSH);
    bglib_output=output;

    // disconnect if connected
    ble_cmd_connection_disconnect(0);
    read_response(7);
    // Stop advertising
    ble_cmd_gap_set_mode(0, 0);
    read_response(6);

    //stop previous operation
    ble_cmd_gap_end_procedure();
    read_response(6);

    // Set tx power: max
    ble_cmd_hardware_set_txpower(15);
    read_response(4);

    // Set advertisement timeout to 20 to 30 ms, advertise on all channels
    ble_cmd_gap_set_adv_parameters(32,48,7);
    read_response(6);

    // Start advertising
    ble_cmd_gap_set_mode(1, 1);
    read_response(6);

// Set current time
    get_time(&adv_time);
    set_adv_time(cts_handle,&adv_time);

    disconnect_issued = 0;
    while(1) {
	status = read_message();
        if(status<0) {
            printf("Error reading message\n");
        }
// Check to see if we need to update the time
	get_time(&current_time);
	if( ( adv_time.tm_year != current_time.tm_year ) ||
	    ( adv_time.tm_mon != current_time.tm_mon ) ||
	    ( adv_time.tm_mday != current_time.tm_mday ) ||
	    ( adv_time.tm_hour != current_time.tm_hour ) ||
	    ( adv_time.tm_min != current_time.tm_min ) ||
	    ( adv_time.tm_sec != current_time.tm_sec ) ) {
#ifdef DEBUG
		printf( "Set time\n" );
#endif
		memcpy(&adv_time,&current_time,sizeof(struct tm));
    		set_adv_time(cts_handle,&adv_time);
	}
#ifdef DEBUG
	printf("Loop\n");
#endif
    }

   return 0;
}
